//Copyright 2015 Red Hat Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// File taken from the cobra project then customized

package main

import (
	"bytes"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"sort"
	"strings"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
)

func prefixLines(s, prefix string) string {
	lines := strings.Split(s, "\n")
	for i := range lines {
		lines[i] = prefix + lines[i]
	}
	return strings.Join(lines, "\n")
}

func printOptionSphinx(buf *bytes.Buffer, cmd *cobra.Command) func(flag *pflag.Flag) {
	return func(flag *pflag.Flag) {
		buf.WriteString(fmt.Sprintf(
			".. option:: --%s\n\n%s\n\n",
			flag.Name,
			prefixLines(flag.Usage, "    ")))
	}
}

func printOptionsSphinx(buf *bytes.Buffer, cmd *cobra.Command, name string) error {
	flags := cmd.NonInheritedFlags()
	flags.SetOutput(buf)
	if flags.HasAvailableFlags() {
		buf.WriteString("Options\n")
		buf.WriteString("~~~~~~~\n\n")
		flags.VisitAll(printOptionSphinx(buf, cmd))
		buf.WriteString("\n")
	}

	parentFlags := cmd.InheritedFlags()
	parentFlags.SetOutput(buf)
	if parentFlags.HasAvailableFlags() {
		buf.WriteString("Options inherited from parent commands\n")
		buf.WriteString("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n")
		parentFlags.VisitAll(printOptionSphinx(buf, cmd))
		buf.WriteString("\n")
	}
	return nil
}

func printSubcommands(buf *bytes.Buffer, cmd *cobra.Command, name string) error {
	children := cmd.Commands()

	if len(children) == 0 {
		return nil
	}

	buf.WriteString("Sub-commands\n")
	buf.WriteString("~~~~~~~~~~~~\n\n")
	buf.WriteString(".. toctree::\n")
	buf.WriteString("    :maxdepth: 1\n\n")
	sort.Sort(byName(children))
	for _, child := range children {
		if !child.IsAvailableCommand() || child.IsAdditionalHelpTopicCommand() {
			continue
		}
		cname := name + " " + child.Name()
		ref := strings.Replace(cname, " ", "_", -1)
		buf.WriteString(fmt.Sprintf("    %s\n", ref))
	}
	buf.WriteString("\n")

	return nil
}

// linkHandler for default Sphinx hyperlink markup
func defaultLinkHandler(name, ref string) string {
	return fmt.Sprintf(":ref:`%s <%s>`", name, ref)
}

// GenSphinx creates reStructured Text output.
func GenSphinx(cmd *cobra.Command, w io.Writer) error {
	return GenSphinxCustom(cmd, w, defaultLinkHandler)
}

// GenSphinxCustom creates custom reStructured Text output.
func GenSphinxCustom(cmd *cobra.Command, w io.Writer, linkHandler func(string, string) string) error {
	cmd.InitDefaultHelpCmd()
	cmd.InitDefaultHelpFlag()

	buf := new(bytes.Buffer)
	name := cmd.CommandPath()

	short := cmd.Short
	long := cmd.Long
	if len(long) == 0 {
		long = short
	}
	ref := "program-" + strings.Replace(name, " ", "_", -1)

	buf.WriteString(".. program:: " + name + "\n\n")
	buf.WriteString(".. _" + ref + ":\n\n")
	title := name + " - " + short
	buf.WriteString(title + "\n")
	buf.WriteString(strings.Repeat("-", len(title)) + "\n\n")
	buf.WriteString("Synopsis\n")
	buf.WriteString("~~~~~~~~\n\n")
	buf.WriteString("\n" + long + "\n\n")

	if cmd.Runnable() {
		buf.WriteString(fmt.Sprintf("::\n\n  %s\n\n", cmd.UseLine()))
	}

	if len(cmd.Example) > 0 {
		buf.WriteString("Examples\n")
		buf.WriteString("~~~~~~~~\n\n")
		buf.WriteString(fmt.Sprintf("::\n\n%s\n\n", indentString(cmd.Example, "  ")))
	}

	if err := printSubcommands(buf, cmd, name); err != nil {
		return err
	}
	if err := printOptionsSphinx(buf, cmd, name); err != nil {
		return err
	}
	if cmd.HasParent() {
		buf.WriteString("SEE ALSO\n")
		buf.WriteString("~~~~~~~~\n\n")
		parent := cmd.Parent()
		pname := parent.CommandPath()
		ref = "program-" + strings.Replace(pname, " ", "_", -1)
		buf.WriteString(fmt.Sprintf("* %s \t - %s\n", linkHandler(pname, ref), parent.Short))
		cmd.VisitParents(func(c *cobra.Command) {
			if c.DisableAutoGenTag {
				cmd.DisableAutoGenTag = c.DisableAutoGenTag
			}
		})
		buf.WriteString("\n")

	}
	if !cmd.DisableAutoGenTag {
		buf.WriteString("*Auto generated by spf13/cobra on " + time.Now().Format("2-Jan-2006") + "*\n")
	}
	_, err := buf.WriteTo(w)
	return err
}

// GenSphinxTree will generate a Sphinx page for this command and all
// descendants in the directory given.
// This function may not work correctly if your command names have `-` in them.
// If you have `cmd` with two subcmds, `sub` and `sub-third`,
// and `sub` has a subcommand called `third`, it is undefined which
// help output will be in the file `cmd-sub-third.1`.
func GenSphinxTree(cmd *cobra.Command, dir string) error {
	emptyStr := func(s string) string { return "" }
	return GenSphinxTreeCustom(cmd, dir, emptyStr, defaultLinkHandler)
}

// GenSphinxTreeCustom is the the same as GenSphinxTree, but
// with custom filePrepender and linkHandler.
func GenSphinxTreeCustom(cmd *cobra.Command, dir string, filePrepender func(string) string, linkHandler func(string, string) string) error {
	for _, c := range cmd.Commands() {
		if !c.IsAvailableCommand() || c.IsAdditionalHelpTopicCommand() {
			continue
		}
		if err := GenSphinxTreeCustom(c, dir, filePrepender, linkHandler); err != nil {
			return err
		}
	}

	basename := strings.Replace(cmd.CommandPath(), " ", "_", -1) + ".rst"
	filename := filepath.Join(dir, basename)
	f, err := os.Create(filename)
	if err != nil {
		return err
	}
	defer f.Close()

	if _, err := io.WriteString(f, filePrepender(filename)); err != nil {
		return err
	}
	if err := GenSphinxCustom(cmd, f, linkHandler); err != nil {
		return err
	}
	return nil
}

// adapted from: https://github.com/kr/text/blob/main/indent.go
func indentString(s, p string) string {
	var res []byte
	b := []byte(s)
	prefix := []byte(p)
	bol := true
	for _, c := range b {
		if bol && c != '\n' {
			res = append(res, prefix...)
		}
		res = append(res, c)
		bol = c == '\n'
	}
	return string(res)
}
